/*
  Copyright (c) 2012, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#define _GNU_SOURCE

#include <errno.h>
#include <libgen.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>

#include <linux/rtnetlink.h>

#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_arp.h>

#include <netinet/in.h>

#include <sys/types.h>
#include <sys/stat.h>


#define ANNOUNCE_INTERVAL 10000
#define ANNOUNCE_TIMEOUT 30000

#define SYSFS_PATH_MAX 256
#define SYSFS_CLASS_NET "/sys/class/net"

#define FFD_MAGIC 0xffd
#define FFVISD_VERSION 0

#define FFD_PORT 0xffd


static const struct in6_addr ffd_addr = { .s6_addr = {0xff, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
						      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0f, 0xfd} };


typedef struct __attribute__((__packed__)) _ffvisd_eth_address_t {
	uint8_t address[ETH_ALEN];
} ffvisd_eth_address_t;

typedef struct __attribute__((__packed__)) _ffvisd_ipv6_host_address_t {
	uint8_t address[8];
} ffvisd_ipv6_host_address_t;

typedef struct __attribute__((__packed__)) _ffvisd_packet_server_announce_t {
	uint32_t seqno;
	ffvisd_eth_address_t eth_address;
	ffvisd_ipv6_host_address_t address;
} ffvisd_packet_server_announce_t;

typedef struct __attribute__((__packed__)) _ffvisd_packet_announce_t {
	uint16_t magic;
	uint16_t version;
	uint32_t n_servers;
	ffvisd_packet_server_announce_t servers[];
} ffvisd_packet_announce_t;

typedef enum _ffvisd_iftype_t {
	IF_UNKNOWN = 0,
	IF_WIRED,
	IF_WIRELESS,
	IF_VIRTUAL,
	IF_BRIDGE,
	IF_MESH,
	IF_MAX
} ffvisd_iftype_t;

typedef struct _ffvisd_iface_t {
	unsigned ifindex;
	ffvisd_iftype_t iftype;
	char *ifname;
	char *bridge_iface;
	char *mesh_iface;
} ffvisd_iface_t;

typedef struct _ffvisd_orig_t {
	struct _ffvisd_orig_t *next;

	ffvisd_eth_address_t eth_address;
} ffvisd_orig_t;

typedef struct _ffvisd_announce_t {
	struct _ffvisd_announce_t *next;

	struct timespec received;
	ffvisd_packet_server_announce_t announce;
} ffvisd_announce_t;

static const char *ffvisd_iftype_names[IF_MAX] = {
	[IF_UNKNOWN] = "unknown",
	[IF_WIRED] = "wired",
	[IF_WIRELESS] = "wireless",
	[IF_VIRTUAL] = "virtual",
	[IF_BRIDGE] = "bridge",
	[IF_MESH] = "mesh",
};


typedef void (*iface_cb)(const ffvisd_iface_t *iface, void *arg);


static int sockfd;
static char *mesh = "bat0";


ffvisd_announce_t *announcements = NULL;


static void ffvisd_iface_free(ffvisd_iface_t *iface);



/* returns (tp1 - tp2) in milliseconds  */
static inline int timespec_diff(const struct timespec *tp1, const struct timespec *tp2) {
	return ((tp1->tv_sec - tp2->tv_sec))*1000 + (tp1->tv_nsec - tp2->tv_nsec)/1e6;
}


static bool file_readv(const char *file, const char *format, va_list ap) {
	FILE *f = fopen(file, "r");
	if (!f)
		return false;

	int ret = vfscanf(f, format, ap);

	fclose(f);

	return (ret > 0);
}

static bool iface_file_read(const char *ifname, const char *file, const char *format, ...) {
	char filename[SYSFS_PATH_MAX];

	snprintf(filename, SYSFS_PATH_MAX, SYSFS_CLASS_NET"/%s/%s", ifname, file);

	va_list ap;
	va_start(ap, format);
	bool ret = file_readv(filename, format, ap);
	va_end(ap);

	return ret;
}

static bool iface_file_exists(const char *ifname, const char *file) {
	char filename[SYSFS_PATH_MAX];

	snprintf(filename, SYSFS_PATH_MAX, SYSFS_CLASS_NET"/%s/%s", ifname, file);

	struct stat st;
	return (stat(filename, &st) == 0);
}

static bool iface_is_mesh(const char *ifname) {
	return iface_file_exists(ifname, "mesh");
}

static char* iface_get_bridge(const char *ifname) {
	char filename[SYSFS_PATH_MAX], filename2[SYSFS_PATH_MAX] = {0};

	snprintf(filename, SYSFS_PATH_MAX, SYSFS_CLASS_NET"/%s/brport/bridge", ifname);
	if (readlink(filename, filename2, sizeof(filename2)) < 0)
		return NULL;

	return strdup(basename(filename2));
}

static struct in6_addr* iface_get_address(ffvisd_iface_t *iface, int scope) {
	struct msg {
		struct nlmsghdr nh;
		struct ifaddrmsg addr;
		uint8_t attrbuf[16384];
	} msg;

	memset(&msg, 0, sizeof(msg));

	int rtnetlink_sk = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);

	msg.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
	msg.nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_ROOT;
	msg.nh.nlmsg_type = RTM_GETADDR;

	msg.addr.ifa_family = AF_INET6;

	if(write(rtnetlink_sk, &msg, msg.nh.nlmsg_len) <= 0) {
		fprintf(stderr, "error: write: %m\n");
		close(rtnetlink_sk);
		return NULL;
	}

	int readlen = read(rtnetlink_sk, &msg, sizeof(msg));

	if (readlen <= 0) {
		fprintf(stderr, "error: read: %m\n");
		close(rtnetlink_sk);
		return NULL;
	}

	close(rtnetlink_sk);

	struct msg *chunk;

	for (chunk = &msg; readlen > sizeof(struct nlmsghdr);) {
		int len = chunk->nh.nlmsg_len - sizeof(struct nlmsghdr);

		if (len < sizeof(struct ifaddrmsg) || readlen < len)
			return NULL;

		if (!NLMSG_OK(&chunk->nh, readlen))
			return NULL;

		if (chunk->nh.nlmsg_type == RTM_NEWADDR && chunk->addr.ifa_scope == scope && chunk->addr.ifa_index == iface->ifindex) {
			struct rtattr *rta = (struct rtattr *)IFA_RTA(&chunk->addr);
			int rtattrlen = IFA_PAYLOAD(&chunk->nh);

			for (; RTA_OK(rta, rtattrlen); rta = RTA_NEXT(rta, rtattrlen)) {
				if(rta->rta_type == IFA_ADDRESS && rta->rta_len == RTA_LENGTH(16)) {
					struct in6_addr *ret = malloc(sizeof(struct in6_addr));
					memcpy(ret->s6_addr, RTA_DATA(rta), 16);
					return ret;
				}
			}
		}

		readlen -= NLMSG_ALIGN(chunk->nh.nlmsg_len);
		chunk = (struct msg*)((uint8_t*)chunk + NLMSG_ALIGN(chunk->nh.nlmsg_len));
	}

	return NULL;
}

static ffvisd_iface_t* ffvisd_iface_new(const char *ifname, unsigned ifindex) {
	if (!ifname || !ifindex)
		return false;

	ffvisd_iface_t *iface = calloc(1, sizeof(ffvisd_iface_t));

	iface->ifname = strdup(ifname);
	iface->ifindex = ifindex;

	if (!iface_file_read(ifname, "batman_adv/mesh_iface", "%as", &iface->mesh_iface) || !strcmp(iface->mesh_iface, "none")) {
		free(iface->mesh_iface);
		iface->mesh_iface = NULL;
	}

	iface->bridge_iface = iface_get_bridge(ifname);

	int type;
	if (iface_file_read(ifname, "type", "%i", &type) && type == ARPHRD_ETHER) {
		if(iface_file_exists(ifname, "mesh"))
			iface->iftype = IF_MESH;
		else if(iface_file_exists(ifname, "bridge"))
			iface->iftype = IF_BRIDGE;
		else if(iface_file_exists(ifname, "tun_flags"))
			iface->iftype = IF_VIRTUAL;
		else if(iface_file_exists(ifname, "wireless"))
			iface->iftype = IF_WIRELESS;
		else
			iface->iftype = IF_WIRED;
	}

	return iface;
}

static void ffvisd_iface_free(ffvisd_iface_t *iface) {
	if (iface) {
		free(iface->ifname);
		free(iface->bridge_iface);
		free(iface->mesh_iface);
		free(iface);
	}
}

static void ffvisd_iface_foreach(iface_cb cb, void *arg) {
	struct if_nameindex *ifaces = if_nameindex();

	if (!ifaces) {
		fprintf(stderr, "error: if_nameindex: %m\n");
		return;
	}

	int i;
	for (i = 0; ifaces[i].if_name; i++) {
		ffvisd_iface_t *iface = ffvisd_iface_new(ifaces[i].if_name, ifaces[i].if_index);

		if (iface)
			cb(iface, arg);

		ffvisd_iface_free(iface);
	}

	if_freenameindex(ifaces);
}


static void pr_iface_info(const ffvisd_iface_t *iface, void *arg) {
	printf("Interface %s: ifindex=%i bridge_iface=%s mesh_iface=%s iftype=%s\n", iface->ifname, iface->ifindex, iface->bridge_iface, iface->mesh_iface, ffvisd_iftype_names[iface->iftype]);
}


static void join_mcast(const ffvisd_iface_t *iface, void *arg) {
	if (!iface->mesh_iface || strcmp(iface->mesh_iface, mesh))
		return;

	struct ipv6_mreq imr;
	imr.ipv6mr_multiaddr = ffd_addr;
	imr.ipv6mr_interface = iface->ifindex;
	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &imr, sizeof(struct ipv6_mreq)) && errno != EADDRINUSE)
		fprintf(stderr, "warning: setsockopt: %m\n");
}

static void pr_announcement(const ffvisd_announce_t *announce) {
	if (!announce->received.tv_sec) {
		printf("local, ");
	}
	else {
		struct timespec tv;
		clock_gettime(CLOCK_MONOTONIC, &tv);
		printf("age: %.1fs, ", timespec_diff(&tv, &announce->received)/1000.0);
	}

	const uint8_t *e = announce->announce.eth_address.address, *a = announce->announce.address.address;

	printf("seqno: %u, ether: %02x:%02x:%02x:%02x:%02x:%02x, address: fe80::%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
	       announce->announce.seqno,
	       e[0], e[1], e[2], e[3], e[4], e[5],
	       a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
}

static bool check_config() {
	if (!iface_is_mesh(mesh)) {
		fprintf(stderr, "error: configured interface is no mesh\n");
		return false;
	}

	return true;
}

static bool init_socket() {
	sockfd = socket(AF_INET6, SOCK_DGRAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "error: socket: %m\n");
		return false;
	}

	int val = 1;
	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_V6ONLY, &val, sizeof(val))) {
		fprintf(stderr, "error: setsockopt: %m\n");
		return false;
	}

	val = 0;
	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &val, sizeof(val))) {
		fprintf(stderr, "error: setsockopt: %m\n");
		return false;
	}

	struct sockaddr_in6 sa6;
	memset(&sa6, 0, sizeof(sa6));
	sa6.sin6_family = AF_INET6;
	sa6.sin6_port = htons(FFD_PORT);

	if (bind(sockfd, (struct sockaddr*)&sa6, sizeof(struct sockaddr_in6))) {
		fprintf(stderr, "error: bind: %m\n");
		return false;
	}

	return true;
}

static bool init_announcements() {
	ffvisd_iface_t *iface;

	char *ifname = iface_get_bridge(mesh);

	if (ifname) {
		iface = ffvisd_iface_new(ifname, if_nametoindex(ifname));
		free(ifname);
	}
	else {
		iface = ffvisd_iface_new(mesh, if_nametoindex(mesh));
	}

	if (!iface)
		return false;

	ffvisd_announce_t *announce = announcements = calloc(1, sizeof(ffvisd_announce_t));

	uint8_t *a = announce->announce.eth_address.address;
	if (!iface_file_read(iface->ifname, "address", "%x:%x:%x:%x:%x:%x", &a[0], &a[1], &a[2], &a[3], &a[4], &a[5]))
		return false;

	struct in6_addr *addr = iface_get_address(iface, RT_SCOPE_LINK);
	if (!addr) {
		ffvisd_iface_free(iface);
		return false;
	}

	memcpy(announce->announce.address.address, addr->s6_addr+8, 8);
	free(addr);

	ffvisd_iface_free(iface);

	return true;
}

static void maintenance() {
	ffvisd_announce_t **announce;

	for(announce = &announcements; *announce; announce = &(*announce)->next) {
		if (!(*announce)->received.tv_sec) /* announcement is local */
			(*announce)->announce.seqno++;
	}
}

int main() {
	if (!check_config())
		return 1;

	ffvisd_iface_foreach(pr_iface_info, NULL);
	puts("");

	if (!init_socket())
		return 1;

	if (!init_announcements())
		return 1;

	while (true) {
		ffvisd_iface_foreach(join_mcast, NULL);

		maintenance();

		puts("Dumping announcements...");
		ffvisd_announce_t *announce;
		for(announce = announcements; announce; announce = announce->next)
			pr_announcement(announce);
		puts("");

		usleep(ANNOUNCE_INTERVAL*1000);
	}

	return 0;
}
